package com.loafle.commons.server;

import java.net.SocketAddress;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.ServerChannel;
import io.netty.channel.nio.NioEventLoopGroup;

/**
 * Server
 */
public abstract class Server {
  private static final Logger logger = LoggerFactory.getLogger(Server.class);
  public static final String CHANNEL_CLASS = "SERVER_CHANNEL_CLASS";
  public static final String CHANNEL_OPTIONS = "SERVER_CHANNEL_OPTIONS";
  public static final String CHANNEL_HANDLER = "SERVER_CHANNEL_HANDLER";
  public static final String CHANNEL_INITIALIZER = "SERVER_CHANNEL_INITIALIZER";
  public static final String SOCKET_ADDRESS = "SERVER_SOCKET_ADDRESS";

  @Value("${server.netty.thread.count.boss}")
  protected int threadCountBoss;
  @Value("${server.netty.thread.count.worker}")
  protected int threadCountWorker;

  @Autowired
  @Qualifier(CHANNEL_CLASS)
  protected Class<? extends ServerChannel> channelClass;
  @Autowired
  @Qualifier(CHANNEL_OPTIONS)
  protected List<ChannelOptionItem<?>> channelOptions;
  @Autowired
  @Qualifier(CHANNEL_HANDLER)
  protected ChannelHandler channelHandler;
  @Autowired
  @Qualifier(CHANNEL_INITIALIZER)
  protected ChannelInitializer<?> channelInitializer;
  @Autowired
  @Qualifier(SOCKET_ADDRESS)
  protected SocketAddress address;

  public void start() throws Exception {
    EventLoopGroup bossGroup = new NioEventLoopGroup(this.threadCountBoss);
    EventLoopGroup workerGroup = new NioEventLoopGroup(this.threadCountWorker);

    try {
      ServerBootstrap serverBootstrap = new ServerBootstrap();
      serverBootstrap.group(bossGroup, workerGroup);

      if (null != this.channelClass) {
          serverBootstrap = serverBootstrap.channel(this.channelClass);
      }
      if (null != this.channelOptions) {
        for (ChannelOptionItem<?> optionItem : this.channelOptions) {
          optionItem.setOption(serverBootstrap);
        }
      }
      if (null != this.channelHandler) {
        serverBootstrap.handler(this.channelHandler);
      }
      serverBootstrap.childHandler(this.channelInitializer);

      this.init();

      ChannelFuture cf = this.bind(serverBootstrap);
      this.onStart();

      cf.channel().closeFuture().sync();

      this.onStop();
    } catch (Exception e) {
      logger.error("Server", e);
    } finally {
      bossGroup.shutdownGracefully();
      workerGroup.shutdownGracefully();
      this.destroy();
    }
  }

  protected ChannelFuture bind(ServerBootstrap serverBootstrap) throws Exception {
    return serverBootstrap.bind(address).sync();
  }

  protected void init() throws Exception {
    // no op
  }
  protected void onStart() throws Exception {
    // no op
  }
  protected void onStop() throws Exception {
    // no op
  }
  protected void destroy() throws Exception {
    // no op
  }

  public static class ChannelOptionItem<T> {
    private final ChannelOption<T> option;
    private final T value;

    public ChannelOptionItem(ChannelOption<T> option, T value) {
      this.option = option;
      this.value = value;
    }
    void setOption(ServerBootstrap serverBootstrap) {
      serverBootstrap.option(option, value);
    }
  }
}