package com.loafle.commons.server.socket.handler.codec;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.util.CharsetUtil;

/**
 * Socket continuation frame containing continuation text or binary data. This is used for
 * fragmented messages where the contents of a messages is contained more than 1 frame.
 */
public class ContinuationSocketFrame extends SocketFrame {

  /**
   * Creates a new empty continuation frame.
   */
  public ContinuationSocketFrame() {
    this(Unpooled.buffer(0));
  }

  /**
   * Creates a new continuation frame with the specified binary data. The final fragment flag is
   * set to true.
   *
   * @param binaryData the content of the frame.
   */
  public ContinuationSocketFrame(ByteBuf binaryData) {
    super(binaryData);
  }

  /**
   * Creates a new continuation frame with the specified binary data
   *
   * @param finalFragment
   *            flag indicating if this frame is the final fragment
   * @param rsv
   *            reserved bits used for protocol extensions
   * @param binaryData
   *            the content of the frame.
   */
  public ContinuationSocketFrame(boolean finalFragment, int rsv, ByteBuf binaryData) {
    super(finalFragment, rsv, binaryData);
  }

  /**
   * Creates a new continuation frame with the specified text data
   *
   * @param finalFragment
   *            flag indicating if this frame is the final fragment
   * @param rsv
   *            reserved bits used for protocol extensions
   * @param text
   *            text content of the frame.
   */
  public ContinuationSocketFrame(boolean finalFragment, int rsv, String text) {
    this(finalFragment, rsv, fromText(text));
  }

  /**
   * Returns the text data in this frame
   */
  public String text() {
    return content().toString(CharsetUtil.UTF_8);
  }

  /**
   * Sets the string for this frame
   *
   * @param text
   *            text to store
   */
  private static ByteBuf fromText(String text) {
    if (text == null || text.isEmpty()) {
      return Unpooled.EMPTY_BUFFER;
    } else {
      return Unpooled.copiedBuffer(text, CharsetUtil.UTF_8);
    }
  }

  @Override
  public ContinuationSocketFrame copy() {
    return (ContinuationSocketFrame) super.copy();
  }

  @Override
  public ContinuationSocketFrame duplicate() {
    return (ContinuationSocketFrame) super.duplicate();
  }

  @Override
  public ContinuationSocketFrame retainedDuplicate() {
    return (ContinuationSocketFrame) super.retainedDuplicate();
  }

  @Override
  public ContinuationSocketFrame replace(ByteBuf content) {
    return new ContinuationSocketFrame(isFinalFragment(), rsv(), content);
  }

  @Override
  public ContinuationSocketFrame retain() {
    super.retain();
    return this;
  }

  @Override
  public ContinuationSocketFrame retain(int increment) {
    super.retain(increment);
    return this;
  }

  @Override
  public ContinuationSocketFrame touch() {
    super.touch();
    return this;
  }

  @Override
  public ContinuationSocketFrame touch(Object hint) {
    super.touch(hint);
    return this;
  }
}