package com.loafle.commons.server.socket.handler.codec;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.util.CharsetUtil;

/**
 * Socket text frame
 */
public class TextSocketFrame extends SocketFrame {

  /**
   * Creates a new empty text frame.
   */
  public TextSocketFrame() {
    super(Unpooled.buffer(0));
  }

  /**
   * Creates a new text frame with the specified text string. The final fragment flag is set to true.
   *
   * @param text
   *            String to put in the frame
   */
  public TextSocketFrame(String text) {
    super(fromText(text));
  }

  /**
   * Creates a new text frame with the specified binary data. The final fragment flag is set to true.
   *
   * @param binaryData
   *            the content of the frame.
   */
  public TextSocketFrame(ByteBuf binaryData) {
    super(binaryData);
  }

  /**
   * Creates a new text frame with the specified text string. The final fragment flag is set to true.
   *
   * @param finalFragment
   *            flag indicating if this frame is the final fragment
   * @param rsv
   *            reserved bits used for protocol extensions
   * @param text
   *            String to put in the frame
   */
  public TextSocketFrame(boolean finalFragment, int rsv, String text) {
    super(finalFragment, rsv, fromText(text));
  }

  private static ByteBuf fromText(String text) {
    if (text == null || text.isEmpty()) {
      return Unpooled.EMPTY_BUFFER;
    } else {
      return Unpooled.copiedBuffer(text, CharsetUtil.UTF_8);
    }
  }

  /**
   * Creates a new text frame with the specified binary data. The final fragment flag is set to true.
   *
   * @param finalFragment
   *            flag indicating if this frame is the final fragment
   * @param rsv
   *            reserved bits used for protocol extensions
   * @param binaryData
   *            the content of the frame.
   */
  public TextSocketFrame(boolean finalFragment, int rsv, ByteBuf binaryData) {
    super(finalFragment, rsv, binaryData);
  }

  /**
   * Returns the text data in this frame
   */
  public String text() {
    return content().toString(CharsetUtil.UTF_8);
  }

  @Override
  public TextSocketFrame copy() {
    return (TextSocketFrame) super.copy();
  }

  @Override
  public TextSocketFrame duplicate() {
    return (TextSocketFrame) super.duplicate();
  }

  @Override
  public TextSocketFrame retainedDuplicate() {
    return (TextSocketFrame) super.retainedDuplicate();
  }

  @Override
  public TextSocketFrame replace(ByteBuf content) {
    return new TextSocketFrame(isFinalFragment(), rsv(), content);
  }

  @Override
  public TextSocketFrame retain() {
    super.retain();
    return this;
  }

  @Override
  public TextSocketFrame retain(int increment) {
    super.retain(increment);
    return this;
  }

  @Override
  public TextSocketFrame touch() {
    super.touch();
    return this;
  }

  @Override
  public TextSocketFrame touch(Object hint) {
    super.touch(hint);
    return this;
  }
}